from types import TracebackType
from typing import Self

import httpx
from httpx_socks import SyncProxyTransport
from solana.rpc.api import Client
from solana.rpc.commitment import Commitment
from solana.rpc.providers.core import _after_request_unparsed  # noqa: RUF100 # noqa
from solana.rpc.providers.http import HTTPProvider
from solders.rpc.requests import Body


class ProxyClient(Client):
    def __init__(
        self,
        endpoint: str | None = None,
        commitment: Commitment | None = None,
        timeout: float = 10,
        extra_headers: dict[str, str] | None = None,
        *,
        proxy: str | None = None,
        **kwargs: dict[str, object],
    ) -> None:
        """Init API client."""
        super().__init__(commitment)
        self._provider: ProxyHTTPProvider = ProxyHTTPProvider(
            endpoint=endpoint, timeout=timeout, extra_headers=extra_headers, proxy=proxy, **kwargs
        )

    def __enter__(self) -> Self:
        self._provider.__enter__()
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None = None,
        exc_value: BaseException | None = None,
        traceback: TracebackType | None = None,
    ) -> None:
        self._provider.__exit__(exc_type, exc_value, traceback)

    def close(self) -> None:
        self._provider.close()


class ProxyHTTPProvider(HTTPProvider):
    def __init__(
        self,
        endpoint: str | None = None,
        extra_headers: dict[str, str] | None = None,
        timeout: float = 10,
        *,
        proxy: str | None = None,
        **kwargs: dict[str, object],
    ) -> None:
        super().__init__(endpoint=endpoint, extra_headers=extra_headers, timeout=timeout, **kwargs)
        self.session = Session(proxy, timeout=timeout)
        self._proxy = proxy

    def __str__(self) -> str:
        return f"HTTP RPC connection {self.endpoint_uri} | proxy={self._proxy}"

    def make_request_unparsed(self, body: Body) -> str:
        request_kwargs = self._before_request(body=body)
        raw_response = self.session.post(**request_kwargs)
        return _after_request_unparsed(raw_response)

    def make_batch_request_unparsed(self, reqs: tuple[Body, ...]) -> str:
        """Make an async HTTP request to an http rpc endpoint."""
        request_kwargs = self._before_batch_request(reqs)
        raw_response = self.session.post(**request_kwargs)
        return _after_request_unparsed(raw_response)

    def is_connected(self) -> bool:
        """Health check."""
        try:
            response = self.session.get(self.health_uri)
            response.raise_for_status()
        except (OSError, httpx.HTTPError) as err:
            self.logger.error("Health check failed with error: %s", str(err))
            return False

        return response.status_code == httpx.codes.OK

    def __enter__(self) -> Self:
        self.session.__enter__()
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None = None,
        exc_value: BaseException | None = None,
        traceback: TracebackType | None = None,
    ) -> None:
        self.session.__exit__(exc_type, exc_value, traceback)

    def close(self) -> None:
        self.session.close()


class Session(httpx.Client):
    def __init__(
        self, proxy: str | None = None, follow_redirects: bool = True, timeout: float = 10, **kwargs: dict[str, object]
    ) -> None:
        if proxy:
            if proxy.startswith("http"):
                kwargs["proxy"] = proxy  # type: ignore[assignment]
            elif proxy.startswith("socks"):
                kwargs["transport"] = SyncProxyTransport.from_url(proxy)
            else:
                raise ValueError(f"Unsupported proxy type: {proxy}")
        # super().__init__(follow_redirects=follow_redirects, timeout=Timeout(timeout), **kwargs)  # type: ignore[arg-type]
        super().__init__(follow_redirects=follow_redirects, **kwargs)  # type: ignore[arg-type]


def get_client(
    endpoint: str,
    commitment: Commitment | None = None,
    extra_headers: dict[str, str] | None = None,
    proxy: str | None = None,
    timeout: float = 10,
    **kwargs: dict[str, object],
) -> ProxyClient:
    return ProxyClient(endpoint, timeout=timeout, proxy=proxy, commitment=commitment, extra_headers=extra_headers, **kwargs)



from pathlib import Path

import click


def run(module: str) -> None:
    example_file = Path(Path(__file__).parent.absolute(), "../examples", f"{module}.yml")
    click.echo(example_file.read_text())



import os

import pydash
import pytest
from dotenv import load_dotenv
from mm_std import fatal, hr

from mm_solana import utils

load_dotenv(".env")


@pytest.fixture
def mainnet_node():
    return os.getenv("MAINNET_NODE")


@pytest.fixture
def testnet_node():
    return os.getenv("TESTNET_NODE")


@pytest.fixture
def usdt_token_address():
    return os.getenv("USDT_TOKEN_ADDRESS")


@pytest.fixture
def usdt_owner_address():
    return os.getenv("USDT_OWNER_ADDRESS")


@pytest.fixture
def binance_wallet():
    return "2ojv9BAiHUrvsm9gxDe7fJSzbNZSJcxZvf8dqmWGHG8S"


@pytest.fixture
def proxy() -> str:
    return os.getenv("PROXY")


@pytest.fixture(scope="session")
def proxies() -> list[str]:
    proxies_url = os.getenv("PROXIES_URL")
    if proxies_url:
        res = hr(proxies_url)
        if res.is_error():
            fatal(f"Can't get proxies: {res.error}")
        proxies = [p.strip() for p in res.body.splitlines() if p.strip()]
        return pydash.uniq(proxies)
    return []


@pytest.fixture
def random_proxy(proxies: list[str]) -> str | None:
    return utils.get_proxy(proxies)



import random
from decimal import Decimal

from mm_std import print_console, print_json, str_to_list
from pydantic import StrictStr, field_validator

from mm_solana.cli import helpers
from mm_solana.cli.helpers import BaseCmdConfig
from mm_solana.transfer import transfer_sol


class Config(BaseCmdConfig):
    from_address: StrictStr
    private_key: StrictStr
    recipients: list[StrictStr]
    nodes: list[StrictStr]
    amount: Decimal

    @field_validator("recipients", "nodes", mode="before")
    def to_list_validator(cls, v: list[str] | str | None) -> list[str]:
        return str_to_list(v)

    @property
    def random_node(self) -> str:
        return random.choice(self.nodes)


def run(config_path: str, print_config: bool) -> None:
    config = helpers.read_config(Config, config_path)
    if print_config:
        print_json(config.model_dump())
        exit(0)

    result = {}
    for recipient in config.recipients:
        res = transfer_sol(
            from_address=config.from_address,
            private_key_base58=config.private_key,
            recipient_address=recipient,
            amount_sol=config.amount,
            nodes=config.nodes,
        )
        result[recipient] = res.ok_or_err()
    print_console(result, print_json=True)



import os
import tomllib

import click
import yaml
from click import Context
from jinja2 import Environment, Template, TemplateSyntaxError, meta
from mm_std import print_console
from pydantic import BaseModel, ConfigDict, ValidationError, field_validator
from rich.console import Console
from rich.table import Table

_jinja_env = Environment(autoescape=True)


class BaseCmdConfig(BaseModel):
    @field_validator("*", mode="before")
    def env_template_validator(cls, v: object) -> object:
        return env_validator(v)

    model_config = ConfigDict(extra="forbid")


def env_validator(v: object) -> object:
    if isinstance(v, str):
        try:
            ast = _jinja_env.parse(v)
            envs = meta.find_undeclared_variables(ast)
            if envs:
                data = {}
                for env in envs:
                    if not os.getenv(env):
                        click.secho(f"can't get environment variable {env}", err=True, fg="red")
                        exit(1)
                    data[env] = os.getenv(env)
                template = Template(v)
                return template.render(data)
        except TemplateSyntaxError as err:
            click.secho(f"jinja syntax error: {err!s}", err=True, fg="red")
            click.secho(v)
            exit(1)
    return v


def read_config_file_or_exit(file_path: str) -> dict[str, object]:
    try:
        with open(file_path, "rb") as f:
            if file_path.endswith(".toml"):
                return tomllib.load(f)
            return yaml.full_load(f)  # type:ignore[no-any-return]
    except Exception as err:
        click.secho(f"can't parse config file: {err!s}", fg="red")
        exit(1)


def print_config_and_exit(ctx: Context, config: BaseCmdConfig) -> None:
    if ctx.obj["config"]:
        print_console(config.model_dump(), print_json=True)
        exit(0)


def read_config[T](config_cls: type[T], config_path: str) -> T:
    try:
        with open(config_path) as f:
            config = config_cls(**yaml.full_load(f))
            return config
    except ValidationError as err:
        table = Table(title="config validation errors")
        table.add_column("field")
        table.add_column("message")
        for e in err.errors():
            loc = e["loc"]
            field = str(loc[0]) if len(loc) > 0 else ""
            table.add_row(field, e["msg"])
        console = Console()
        console.print(table)
        exit(1)



import random
from decimal import Decimal
from typing import Any

from mm_std import print_console, print_json, str_to_list
from pydantic import StrictStr, field_validator

from mm_solana import balance, utils
from mm_solana.cli import helpers
from mm_solana.cli.helpers import BaseCmdConfig
from mm_solana.token import get_balance_with_retries


class Config(BaseCmdConfig):
    accounts: list[StrictStr]
    nodes: list[StrictStr]
    tokens: list[StrictStr] | None = None

    @field_validator("accounts", "nodes", "tokens", mode="before")
    def to_list_validator(cls, v: list[str] | str | None) -> list[str]:
        return str_to_list(v)

    @property
    def random_node(self) -> str:
        return random.choice(self.nodes)


def run(config_path: str, print_config: bool) -> None:
    config = helpers.read_config(Config, config_path)
    if print_config:
        print_json(config.model_dump())
        exit(0)

    # config = parse_config(ctx, config_path, Config)
    # print_config_and_exit(ctx, config)
    result: dict[str, Any] = {"sol": _get_sol_balances(config.accounts, config.nodes)}
    result["sol_sum"] = sum([v for v in result["sol"].values() if v is not None])

    if config.tokens:
        for token in config.tokens:
            result[token] = _get_token_balances(token, config.accounts, config.nodes)
            result[token + "_sum"] = sum([v for v in result[token].values() if v is not None])

    print_console(result, print_json=True)


def _get_token_balances(token: str, accounts: list[str], nodes: list[str]) -> dict[str, int | None]:
    result = {}
    for account in accounts:
        # result[account] = _get_token_balance(token, account, nodes)
        result[account] = get_balance_with_retries(
            nodes=nodes,
            owner_address=account,
            token_mint_address=token,
            retries=3,
        ).ok_or_none()
    return result


def _get_sol_balances(accounts: list[str], nodes: list[str]) -> dict[str, Decimal | None]:
    result = {}
    for account in accounts:
        res = balance.get_balance_with_retries(nodes=nodes, address=account, retries=3)
        result[account] = utils.lamports_to_sol(res.unwrap(), ndigits=2) if res.is_ok() else None
    return result



from decimal import Decimal

from mm_solana import utils
from mm_solana.rpc import DEFAULT_MAINNET_RPC


def test_lamports_to_sol():
    res = utils.lamports_to_sol(272356343007, ndigits=4)
    assert res == Decimal("272.3563")


def test_get_node():
    assert utils.get_node() == DEFAULT_MAINNET_RPC
    assert utils.get_node("n1") == "n1"
    res = utils.get_node(["n1", "n2"])
    assert res == "n1" or res == "n2"


def test_get_proxy():
    assert utils.get_proxy([]) is None



import traceback

from mm_std import Err, Ok, Result
from solana.exceptions import SolanaRpcException
from solana.rpc.types import TokenAccountOpts
from solders.pubkey import Pubkey

from mm_solana.client import get_client
from mm_solana.types import Nodes, Proxies
from mm_solana.utils import get_node, get_proxy


def get_balance(
    node: str,
    owner_address: str,
    token_mint_address: str,
    token_account: str | None = None,
    timeout: float = 10,
    proxy: str | None = None,
) -> Result[int]:
    try:
        with get_client(node, proxy=proxy, timeout=timeout) as client:
            if token_account:
                res_balance = client.get_token_account_balance(Pubkey.from_string(token_account))
                return Ok(int(res_balance.value.amount))

            res_accounts = client.get_token_accounts_by_owner(
                Pubkey.from_string(owner_address),
                TokenAccountOpts(mint=Pubkey.from_string(token_mint_address)),
            )
            if not res_accounts.value:
                return Err("no_token_accounts")

            token_accounts = [a.pubkey for a in res_accounts.value]
            balances = []
            for token_account_ in token_accounts:
                res = client.get_token_account_balance(token_account_)
                if res.value:  # type:ignore[truthy-bool]
                    balances.append(int(res.value.amount))

            if len(balances) > 1:
                return Err("there are many non empty token accounts, set token_account explicitly")
            return Ok(balances[0])
    except SolanaRpcException as e:
        return Err(e.error_msg)
    except Exception as e:
        traceback.print_exc()
        return Err(e)


def get_balance_with_retries(
    nodes: Nodes,
    owner_address: str,
    token_mint_address: str,
    retries: int,
    token_account: str | None = None,
    timeout: float = 10,
    proxies: Proxies = None,
) -> Result[int]:
    res: Result[int] = Err("not started yet")
    for _ in range(retries):
        res = get_balance(
            get_node(nodes), owner_address, token_mint_address, token_account, timeout=timeout, proxy=get_proxy(proxies)
        )
        if res.is_ok():
            return res
    return res


def get_decimals(node: str, token_mint_address: str, timeout: float = 10, proxy: str | None = None) -> Result[int]:
    try:
        with get_client(node, proxy=proxy, timeout=timeout) as client:
            res = client.get_token_supply(Pubkey.from_string(token_mint_address))
            return Ok(res.value.decimals)
    except Exception as e:
        return Err(e)


def get_decimals_with_retries(
    nodes: Nodes, token_mint_address: str, retries: int, timeout: float = 10, proxies: Proxies = None
) -> Result[int]:
    res: Result[int] = Err("not started yet")
    for _ in range(retries):
        res = get_decimals(get_node(nodes), token_mint_address, timeout=timeout, proxy=get_proxy(proxies))
        if res.is_ok():
            return res
    return res


# def transfer_to_wallet_address(
#     *,
#     node: str,
#     private_key: str,
#     recipient_wallet_address: str,
#     token_mint_address: str,
#     amount: int,
# ) -> Result[str]:
#     try:
#         keypair = account.get_keypair(private_key)
#         token_client = Token(Client(node), Pubkey.from_string(token_mint_address), program_id=TOKEN_PROGRAM_ID, payer=keypair)
#
#         # get from_token_account
#         res = token_client.get_accounts(keypair.public_key)
#         token_accounts = res["result"]["value"]
#         if len(token_accounts) > 1:
#             return Result(error="many_from_token_accounts", data=res)
#         from_token_account = Pubkey.from_string(token_accounts[0]["pubkey"])
#
#         # get to_token_account
#         res = token_client.get_accounts(Pubkey.from_string(recipient_wallet_address))
#         token_accounts = res["result"]["value"]
#         if len(token_accounts) > 1:
#             return Result(error="many_to_token_accounts", data=res)
#         elif len(token_accounts) == 1:
#             to_token_account = Pubkey.from_string(token_accounts[0]["pubkey"])
#         else:  # create a new to_token_account
#             to_token_account = token_client.create_account(owner=Pubkey.from_string(recipient_wallet_address))
#
#         res = token_client.transfer(source=from_token_account, dest=to_token_account, owner=keypair, amount=amount)
#         if res.get("result"):
#             return Result(ok=res.get("result"), data=res)
#         return Result(error="unknown_response", data=res)
#     except RPCException as e:
#         return Result(error="rcp_exception", data=str(e))
#     except Exception as e:
#         return Result(error="exception", data=str(e))



from typing import Any

from mm_std import Err, Ok, Result, hr
from pydantic import BaseModel, ConfigDict, Field

DEFAULT_MAINNET_RPC = "https://api.mainnet-beta.solana.com"
DEFAULT_TESTNET_RPC = "https://api.testnet.solana.com"


class EpochInfo(BaseModel):
    model_config = ConfigDict(populate_by_name=True)

    epoch: int
    absolute_slot: int = Field(..., alias="absoluteSlot")
    block_height: int = Field(..., alias="blockHeight")
    slot_index: int = Field(..., alias="slotIndex")
    slots_in_epoch: int = Field(..., alias="slotsInEpoch")
    transaction_count: int = Field(..., alias="transactionCount")

    @property
    def progress(self) -> float:
        return round(self.slot_index / self.slots_in_epoch * 100, 2)


class ClusterNode(BaseModel):
    pubkey: str
    version: str | None
    gossip: str | None
    rpc: str | None


class VoteAccount(BaseModel):
    class EpochCredits(BaseModel):
        epoch: int
        credits: int
        previous_credits: int

    validator: str
    vote: str
    commission: int
    stake: int
    credits: list[EpochCredits]
    epoch_vote_account: bool
    root_slot: int
    last_vote: int
    delinquent: bool


class BlockProduction(BaseModel):
    class Leader(BaseModel):
        address: str
        produced: int
        skipped: int

    slot: int
    first_slot: int
    last_slot: int
    leaders: list[Leader]

    @property
    def total_produced(self) -> int:
        return sum(leader.produced for leader in self.leaders)

    @property
    def total_skipped(self) -> int:
        return sum(leader.skipped for leader in self.leaders)


class StakeActivation(BaseModel):
    state: str
    active: int
    inactive: int


def rpc_call(
    *,
    node: str,
    method: str,
    params: list[Any],
    id_: int = 1,
    timeout: int = 10,
    proxy: str | None = None,
) -> Result[Any]:
    data = {"jsonrpc": "2.0", "method": method, "params": params, "id": id_}
    if node.startswith("http"):
        return _http_call(node, data, timeout, proxy)
    raise NotImplementedError("ws is not implemented")


def _http_call(node: str, data: dict[str, object], timeout: int, proxy: str | None) -> Result[Any]:
    res = hr(node, method="POST", proxy=proxy, timeout=timeout, params=data, json_params=True)
    try:
        if res.is_error():
            return res.to_err_result()

        err = res.json.get("error", {}).get("message", "")
        if err:
            return res.to_err_result(f"service_error: {err}")
        if "result" in res.json:
            return res.to_ok_result(res.json["result"])

        return res.to_err_result("unknown_response")
    except Exception as e:
        return res.to_err_result(f"exception: {e!s}")


def get_balance(node: str, address: str, timeout: int = 10, proxy: str | None = None) -> Result[int]:
    """Returns balance in lamports"""
    return rpc_call(node=node, method="getBalance", params=[address], timeout=timeout, proxy=proxy).and_then(lambda r: r["value"])


def get_block_height(node: str, timeout: int = 10, proxy: str | None = None) -> Result[int]:
    """Returns balance in lamports"""
    return rpc_call(node=node, method="getBlockHeight", params=[], timeout=timeout, proxy=proxy)


def get_slot(node: str, timeout: int = 10, proxy: str | None = None) -> Result[int]:
    return rpc_call(node=node, method="getSlot", params=[], timeout=timeout, proxy=proxy)


def get_epoch_info(node: str, epoch: int | None = None, timeout: int = 10, proxy: str | None = None) -> Result[EpochInfo]:
    params = [epoch] if epoch else []
    return rpc_call(node=node, method="getEpochInfo", params=params, timeout=timeout, proxy=proxy).and_then(
        lambda r: EpochInfo(**r),
    )


def get_health(node: str, timeout: int = 10, proxy: str | None = None) -> Result[bool]:
    return rpc_call(node=node, method="getHealth", params=[], timeout=timeout, proxy=proxy).and_then(lambda r: r == "ok")


def get_cluster_nodes(node: str, timeout: int = 30, proxy: str | None = None) -> Result[list[ClusterNode]]:
    return rpc_call(node=node, method="getClusterNodes", timeout=timeout, proxy=proxy, params=[]).and_then(
        lambda r: [ClusterNode(**n) for n in r],
    )


def get_vote_accounts(node: str, timeout: int = 30, proxy: str | None = None) -> Result[list[VoteAccount]]:
    res = rpc_call(node=node, method="getVoteAccounts", timeout=timeout, proxy=proxy, params=[])
    if res.is_err():
        return res
    try:
        data = res.unwrap()
        result: list[VoteAccount] = []
        for a in data["current"]:
            result.append(
                VoteAccount(
                    validator=a["nodePubkey"],
                    vote=a["votePubkey"],
                    commission=a["commission"],
                    stake=a["activatedStake"],
                    credits=[
                        VoteAccount.EpochCredits(epoch=c[0], credits=c[1], previous_credits=c[2]) for c in a["epochCredits"]
                    ],
                    delinquent=False,
                    epoch_vote_account=a["epochVoteAccount"],
                    root_slot=a["rootSlot"],
                    last_vote=a["lastVote"],
                ),
            )
        for a in data["delinquent"]:
            result.append(
                VoteAccount(
                    validator=a["nodePubkey"],
                    vote=a["votePubkey"],
                    commission=a["commission"],
                    stake=a["activatedStake"],
                    credits=[
                        VoteAccount.EpochCredits(epoch=c[0], credits=c[1], previous_credits=c[2]) for c in a["epochCredits"]
                    ],
                    delinquent=True,
                    epoch_vote_account=a["epochVoteAccount"],
                    root_slot=a["rootSlot"],
                    last_vote=a["lastVote"],
                ),
            )
        return Ok(result, res.data)
    except Exception as e:
        return Err(e, res.data)


def get_leader_scheduler(
    node: str,
    slot: int | None = None,
    timeout: int = 10,
    proxy: str | None = None,
) -> Result[dict[str, list[int]]]:
    return rpc_call(
        node=node,
        method="getLeaderSchedule",
        timeout=timeout,
        proxy=proxy,
        params=[slot],
    )


def get_block_production(node: str, timeout: int = 60, proxy: str | None = None) -> Result[BlockProduction]:
    res = rpc_call(node=node, method="getBlockProduction", timeout=timeout, proxy=proxy, params=[])
    if res.is_err():
        return res
    try:
        res_ok = res.unwrap()
        slot = res_ok["context"]["slot"]
        first_slot = res_ok["value"]["range"]["firstSlot"]
        last_slot = res_ok["value"]["range"]["lastSlot"]
        leaders = []
        for address, (leader, produced) in res.ok["value"]["byIdentity"].items():  # type: ignore[index]
            leaders.append(BlockProduction.Leader(address=address, produced=produced, skipped=leader - produced))
        return Ok(BlockProduction(slot=slot, first_slot=first_slot, last_slot=last_slot, leaders=leaders), res.data)
    except Exception as e:
        return Err(e, data=res.data)


def get_stake_activation(node: str, address: str, timeout: int = 60, proxy: str | None = None) -> Result[StakeActivation]:
    return rpc_call(node=node, method="getStakeActivation", timeout=timeout, proxy=proxy, params=[address]).and_then(
        lambda ok: StakeActivation(**ok),
    )


def get_transaction(
    node: str,
    signature: str,
    max_supported_transaction_version: int | None = None,
    encoding: str = "json",
    timeout: int = 60,
    proxy: str | None = None,
) -> Result[dict[str, object] | None]:
    if max_supported_transaction_version is not None:
        params = [signature, {"maxSupportedTransactionVersion": max_supported_transaction_version, "encoding": encoding}]
    else:
        params = [signature, encoding]
    return rpc_call(node=node, method="getTransaction", timeout=timeout, proxy=proxy, params=params)



# import click
# from click import Context
# from click_aliases import ClickAliasedGroup
#
# from mm_solana.cli.cmd import (
#     balance_cmd,
#     example_cmd,
#     generate_accounts_cmd,
#     keypair_cmd,
#     transfer_sol_cmd,
# )
# from mm_solana.cli.helpers import get_version
#
#
# @click.group(cls=ClickAliasedGroup)
# @click.option("-c", "--config/--no-config", "config_", default=False, help="Print config and exit")
# @click.option("-n", "--node", multiple=True, help="List of JSON RPC nodes, it overwrites node/nodes field in config")
# @click.version_option(get_version(), help="Show the version and exit")
# @click.help_option(help="Show this message and exit")
# @click.pass_context
# def cli(ctx: Context, config_: bool, node: list[str]) -> None:
#     ctx.ensure_object(dict)
#     ctx.obj["config"] = config_
#     ctx.obj["nodes"] = node
#
#
# # noinspection PyTypeChecker
# cli.add_command(balance_cmd.cli)
# # noinspection PyTypeChecker
# cli.add_command(example_cmd.cli)
# # noinspection PyTypeChecker
# cli.add_command(generate_accounts_cmd.cli)
# # noinspection PyTypeChecker
# cli.add_command(keypair_cmd.cli)
# # noinspection PyTypeChecker
# cli.add_command(transfer_sol_cmd.cli)



from mm_std import Err, Ok, Result
from pydantic import BaseModel

from mm_solana.rpc import rpc_call


class BlockTxCount(BaseModel):
    slot: int
    block_time: int | None
    vote_tx_ok: int
    vote_tx_error: int
    non_vote_tx_ok: int
    non_vote_tx_error: int


def calc_block_tx_count(node: str, slot: int, timeout: int = 10, proxy: str | None = None) -> Result[BlockTxCount]:
    res = rpc_call(node=node, method="getBlock", params=[slot], timeout=timeout, proxy=proxy)
    if res.is_err():
        return res
    vote_tx_ok = 0
    vote_tx_error = 0
    non_vote_tx_ok = 0
    non_vote_tx_error = 0
    vote_tx_keys = [
        "SysvarS1otHashes111111111111111111111111111",
        "SysvarC1ock11111111111111111111111111111111",
        "Vote111111111111111111111111111111111111111",
    ]
    try:
        res_ok = res.unwrap()
        txs = res_ok["transactions"]
        block_time = res_ok["blockTime"]
        for tx in txs:
            is_error = tx["meta"]["err"] is not None
            account_keys = tx["transaction"]["message"]["accountKeys"]
            if len(account_keys) == 5 and vote_tx_keys == account_keys[2:]:
                if is_error:
                    vote_tx_error += 1
                else:
                    vote_tx_ok += 1
            else:
                if is_error:
                    non_vote_tx_error += 1
                else:
                    non_vote_tx_ok += 1

        return Ok(
            BlockTxCount(
                slot=slot,
                vote_tx_ok=vote_tx_ok,
                vote_tx_error=vote_tx_error,
                non_vote_tx_ok=non_vote_tx_ok,
                non_vote_tx_error=non_vote_tx_error,
                block_time=block_time,
            ),
            res.data,
        )
    except Exception as e:
        return Err(e, data=res.data)



from mm_std import print_console

from mm_solana import rpc


def run(urls: list[str], proxy: str | None) -> None:
    for url in urls:
        res = rpc.get_block_height(url, proxy=proxy, timeout=10)
        print_console(url, res.ok_or_err())



from decimal import Decimal

import pydash
from mm_std import Err, Ok, Result
from pydantic import BaseModel
from solana.rpc.api import Client
from solana.transaction import Transaction
from solders.pubkey import Pubkey
from solders.system_program import TransferParams, transfer

from mm_solana import rpc, utils
from mm_solana.account import check_private_key, get_keypair


def transfer_sol(
    *,
    from_address: str,
    private_key_base58: str,
    recipient_address: str,
    amount_sol: Decimal,
    nodes: str | list[str] | None = None,
    attempts: int = 3,
) -> Result[str]:
    acc = get_keypair(private_key_base58)
    if not check_private_key(from_address, private_key_base58):
        raise ValueError("from_address or private_key_base58 is invalid")

    lamports = int(amount_sol * 10**9)
    error = None
    data = None
    for _ in range(attempts):
        try:
            client = Client(utils.get_node(nodes))
            tx = Transaction(fee_payer=acc.pubkey())
            ti = transfer(
                TransferParams(from_pubkey=acc.pubkey(), to_pubkey=Pubkey.from_string(recipient_address), lamports=lamports),
            )
            tx.add(ti)
            res = client.send_transaction(tx, acc)
            data = res.to_json()
            return Ok(str(res.value), data=data)
        except Exception as e:
            error = e

    return Err(error or "unknown", data=data)


class TransferInfo(BaseModel):
    source: str
    destination: str
    lamports: int


def find_transfers(node: str, tx_signature: str) -> Result[list[TransferInfo]]:
    res = rpc.get_transaction(node, tx_signature, encoding="jsonParsed")
    if res.is_err():
        return res  # type: ignore[return-value]
    result = []
    try:
        for ix in pydash.get(res.ok, "transaction.message.instructions"):
            program_id = ix.get("programId")
            ix_type = pydash.get(ix, "parsed.type")
            if program_id == "11111111111111111111111111111111" and ix_type == "transfer":
                source = pydash.get(ix, "parsed.info.source")
                destination = pydash.get(ix, "parsed.info.destination")
                lamports = pydash.get(ix, "parsed.info.lamports")
                if source and destination and lamports:
                    result.append(TransferInfo(source=source, destination=destination, lamports=lamports))
        return Ok(result, data=res.data)
    except Exception as e:
        return Err(e, res.data)



from pathlib import Path

from mm_std import print_console

from mm_solana.account import (
    get_private_key_arr_str,
    get_private_key_base58,
    get_public_key,
)


def run(private_key: str) -> None:
    if (file := Path(private_key)).is_file():
        private_key = file.read_text()

    public = get_public_key(private_key)
    private_base58 = get_private_key_base58(private_key)
    private_arr = get_private_key_arr_str(private_key)
    print_console({"public": public, "private_base58": private_base58, "private_arr": private_arr}, print_json=True)



import random

import base58
import pydash
from mm_std import Err, Ok, Result
from pydantic import BaseModel
from solana.rpc.api import Client
from solders.keypair import Keypair
from solders.pubkey import Pubkey
from solders.rpc.responses import GetAccountInfoResp


class NewAccount(BaseModel):
    public_key: str
    private_key_base58: str
    private_key_arr: list[int]


def generate_account() -> NewAccount:
    keypair = Keypair()
    public_key = str(keypair.pubkey())
    private_key_base58 = base58.b58encode(bytes(keypair.to_bytes_array())).decode("utf-8")
    private_key_arr = [x for x in keypair.to_bytes_array()]
    return NewAccount(public_key=public_key, private_key_base58=private_key_base58, private_key_arr=private_key_arr)


def get_keypair(private_key: str | list[int]) -> Keypair:
    if isinstance(private_key, str):
        if "[" in private_key:
            private_key_ = [int(x) for x in private_key.replace("[", "").replace("]", "").split(",")]
        else:
            private_key_ = base58.b58decode(private_key)  # type: ignore[assignment]
    else:
        private_key_ = private_key
    return Keypair.from_bytes(private_key_)


def check_private_key(public_key_base58: str, private_key: str | list[int]) -> bool:
    return str(get_keypair(private_key).pubkey()) == public_key_base58


def get_public_key(private_key: str) -> str:
    if "[" in private_key:
        private_key_ = [int(x) for x in private_key.replace("[", "").replace("]", "").split(",")]
    else:
        private_key_ = base58.b58decode(private_key)  # type: ignore[assignment]
    return str(Keypair.from_bytes(private_key_).pubkey())


def get_private_key_base58(private_key: str) -> str:
    keypair = get_keypair(private_key)
    return base58.b58encode(bytes(keypair.to_bytes_array())).decode("utf-8")


def get_private_key_arr(private_key: str) -> list[int]:
    keypair = get_keypair(private_key)
    return list(x for x in keypair.to_bytes_array())  # noqa: C400


def get_private_key_arr_str(private_key: str) -> str:
    return f"[{','.join(str(x) for x in get_private_key_arr(private_key))}]"


def is_empty_account(*, address: str, node: str | None = None, nodes: list[str] | None = None, attempts: int = 3) -> Result[bool]:
    if not node and not nodes:
        raise ValueError("node or nodes must be set")
    error = None
    data = None
    for _ in range(attempts):
        try:
            client = Client(node or random.choice(nodes))  # type: ignore[arg-type]
            res: GetAccountInfoResp = client.get_account_info(Pubkey.from_string(address))
            data = res
            slot = pydash.get(res, "result.context.slot")
            value = pydash.get(res, "result.value")
            if slot and value is None:
                return Ok(True, data=data)
            if slot and value:
                return Ok(False, data=data)
        except Exception as e:
            error = str(e)
    return Err(error or "unknown response", data=data)



from mm_solana import token
from mm_solana.account import generate_account


def test_get_balance(mainnet_node, usdt_token_address, usdt_owner_address, random_proxy):
    # existing token account
    res = token.get_balance(mainnet_node, usdt_owner_address, usdt_token_address, proxy=random_proxy)
    assert res.unwrap() > 0

    res = token.get_balance(mainnet_node, generate_account().public_key, usdt_token_address, proxy=random_proxy)
    assert res.err == "no_token_accounts"


def test_get_decimals(mainnet_node, usdt_token_address, random_proxy):
    res = token.get_decimals(mainnet_node, usdt_token_address, proxy=random_proxy)
    assert res.unwrap() == 6



from mm_solana import rpc


def test_get_balance(mainnet_node, binance_wallet, random_proxy):
    res = rpc.get_balance(mainnet_node, binance_wallet, proxy=random_proxy)
    assert res.unwrap() > 10_000_000


def test_get_block_height(mainnet_node, random_proxy):
    res = rpc.get_block_height(mainnet_node, proxy=random_proxy)
    assert res.unwrap() > 10_000_000


def test_get_slot(testnet_node, random_proxy):
    res = rpc.get_slot(testnet_node, proxy=random_proxy)
    assert res.unwrap() > 10_000


def test_get_epoch_info(testnet_node, random_proxy):
    res = rpc.get_epoch_info(testnet_node, proxy=random_proxy)
    assert res.unwrap().epoch > 500


def test_get_health(mainnet_node, testnet_node, random_proxy):
    res = rpc.get_health(mainnet_node, proxy=random_proxy)
    assert res.unwrap() is True

    res = rpc.get_health(testnet_node, proxy=random_proxy)
    assert res.unwrap() is True


def test_get_cluster_nodes(testnet_node, random_proxy):
    res = rpc.get_cluster_nodes(testnet_node, proxy=random_proxy)
    assert len(res.unwrap()) > 10


def test_get_vote_accounts(testnet_node, random_proxy):
    res = rpc.get_vote_accounts(testnet_node, timeout=60, proxy=random_proxy)
    assert len(res.unwrap()) > 10


def test_get_leader_scheduler(testnet_node, random_proxy):
    res = rpc.get_leader_scheduler(testnet_node, proxy=random_proxy)
    assert len(res.unwrap()) > 10


def test_get_block_production(testnet_node, random_proxy):
    res = rpc.get_block_production(testnet_node, proxy=random_proxy)
    assert res.unwrap().slot > 1000


# def test_get_stake_activation(mainnet_node, random_proxy):
#     res = rpc.get_stake_activation(
#         mainnet_node, "GitYucwpNcg6Dx1Y15UQ9TQn8LZMX1uuqQNn8rXxEWNC", proxy=random_proxy
#     )  # from kraken
#     assert res.unwrap().state == "active"


def test_get_transaction(mainnet_node, random_proxy):
    tx_hash = "2vifJ5g4inS4spZLQMUyVstvMrCM2mg1QC9xjD6bgsiMUwp8sTE5waCdshJ8SVaH95WGtexjH3q8ot1GoKe9yK3h"
    res = rpc.get_transaction(mainnet_node, tx_hash, 0, proxy=random_proxy)
    assert res.unwrap()["blockTime"] == 1708667439



from mm_std import Err, Result

from mm_solana import rpc
from mm_solana.types import Nodes, Proxies
from mm_solana.utils import get_node, get_proxy


def get_balance(node: str, address: str, timeout: int = 10, proxy: str | None = None) -> Result[int]:
    return rpc.get_balance(node, address, timeout, proxy)


def get_balance_with_retries(nodes: Nodes, address: str, retries: int, timeout: int = 10, proxies: Proxies = None) -> Result[int]:
    res: Result[int] = Err("not started yet")
    for _ in range(retries):
        res = get_balance(get_node(nodes), address, timeout=timeout, proxy=get_proxy(proxies))
        if res.is_ok():
            return res
    return res



from mm_solana.balance import get_balance


def test_get_balance(mainnet_node, usdt_owner_address, random_proxy):
    res = get_balance(mainnet_node, usdt_owner_address, proxy=random_proxy)
    assert res.unwrap() > 10



from enum import Enum
from typing import Annotated

import typer
from mm_std import print_plain

from mm_solana.cli.cmd import balance_cmd, example_cmd, generate_accounts_cmd, keypair_cmd, node_cmd, transfer_sol_cmd

app = typer.Typer(no_args_is_help=True, pretty_exceptions_enable=False, add_completion=False)


def version_callback(value: bool) -> None:
    if value:
        import importlib.metadata

        print_plain(f"mm-solana version: {importlib.metadata.version("mm-solana")}")
        raise typer.Exit()


@app.callback()
def main(_version: bool = typer.Option(None, "--version", callback=version_callback, is_eager=True)) -> None:
    pass


class ConfigExample(str, Enum):
    balance = "balance"
    transfer_sol = "transfer-sol"


@app.command(name="example", help="Print an example of config for a command")
def example_command(command: Annotated[ConfigExample, typer.Argument()]) -> None:
    example_cmd.run(command.value)


@app.command(name="generate-accounts", help="Generate new accounts")
def generate_accounts_command(
    limit: Annotated[int, typer.Option("--limit", "-l")] = 5,
    array: Annotated[bool, typer.Option("--array", help="Print private key in the array format.")] = False,
) -> None:
    generate_accounts_cmd.run(limit, array)


@app.command(name="keypair", help="Print public, private_base58, private_arr by a private key")
def keypair_command(private_key: str) -> None:
    keypair_cmd.run(private_key)


@app.command(name="balance", help="Print SOL and tokens balances")
def balance_command(
    config_path: str, print_config: Annotated[bool, typer.Option("--config", "-c", help="Print config and exit")] = False
) -> None:
    balance_cmd.run(config_path, print_config)


@app.command(name="transfer-sol", help="Transfer SOL")
def transfer_sol_command(
    config_path: str, print_config: Annotated[bool, typer.Option("--config", "-c", help="Print config and exit")] = False
) -> None:
    transfer_sol_cmd.run(config_path, print_config)


@app.command(name="node", help="Check RPC urls")
def node_command(
    urls: Annotated[list[str], typer.Argument()],
    proxy: Annotated[str | None, typer.Option("--proxy", "-p", help="Proxy")] = None,
) -> None:
    node_cmd.run(urls, proxy)


if __name__ == "__main_":
    app()



import random
from decimal import Decimal

from mm_solana.rpc import DEFAULT_MAINNET_RPC
from mm_solana.types import Nodes, Proxies


def lamports_to_sol(lamports: int, ndigits: int = 4) -> Decimal:
    return Decimal(str(round(lamports / 10**9, ndigits=ndigits)))


def get_node(nodes: Nodes | None = None) -> str:
    if nodes is None:
        return DEFAULT_MAINNET_RPC
    if isinstance(nodes, str):
        return nodes
    return random.choice(nodes)
    # match nodes:
    #     case None:
    #         return DEFAULT_MAINNET_RPC
    #     case list():
    #         return random.choice(nodes)
    #     case _:
    #         return nodes


def get_proxy(proxies: Proxies) -> str | None:
    if not proxies:
        return None
    if isinstance(proxies, str):
        return proxies
    return random.choice(proxies)

    # match proxies:
    #     case [] | None:
    #         return None
    #     case list():
    #         return random.choice(proxies)
    #     case _:
    #         return proxies



import json
import os
import random
from decimal import Decimal
from typing import Literal

import pydash
from mm_std import CommandResult, Err, Ok, Result, run_command, run_ssh_command
from pydantic import BaseModel, ConfigDict, Field, field_validator


class ValidatorInfo(BaseModel):
    identity_address: str
    info_address: str
    name: str | None
    keybase: str | None
    website: str | None
    details: str | None


class StakeAccount(BaseModel):
    type: str = Field(..., alias="stakeType")
    balance: float | None = Field(..., alias="accountBalance")
    withdrawer: str
    staker: str
    vote: str | None = Field(None, alias="delegatedVoteAccountAddress")

    @field_validator("balance")
    def from_lamports_to_sol(cls, v: int | None) -> float | None:
        if v:
            return v / 1_000_000_000


class Stake(BaseModel):
    model_config = ConfigDict(populate_by_name=True)

    stake_address: str = Field(..., alias="stakePubkey")
    withdrawer_address: str = Field(..., alias="withdrawer")
    vote_address: str | None = Field(None, alias="delegatedVoteAccountAddress")
    balance: float | None = Field(..., alias="accountBalance")
    delegated: float | None = Field(None, alias="delegatedStake")
    active: float | None = Field(None, alias="activeStake")
    lock_time: int | None = Field(None, alias="unixTimestamp")

    @field_validator("balance", "delegated", "active")
    def from_lamports_to_sol(cls, v: int | None) -> float | None:
        if v:
            return v / 1_000_000_000


def get_balance(
    *,
    address: str,
    solana_dir: str = "",
    url: str = "localhost",
    ssh_host: str | None = None,
    ssh_key_path: str | None = None,
    timeout: int = 60,
) -> Result[Decimal]:
    solana_dir = _solana_dir(solana_dir)
    cmd = f"{solana_dir}solana balance {address} -u {url}"
    res = _exec_cmd(cmd, ssh_host, ssh_key_path, timeout)
    data = {"cmd": cmd, "stdout": res.stdout, "stderr": res.stderr}
    try:
        return Ok(Decimal(res.stdout.replace("SOL", "").strip()), data=data)
    except Exception as e:
        return Err(e, data=data)


def get_stake_account(
    *,
    address: str,
    solana_dir: str = "",
    url: str = "localhost",
    ssh_host: str | None = None,
    ssh_key_path: str | None = None,
    timeout: int = 60,
) -> Result[StakeAccount]:
    solana_dir = _solana_dir(solana_dir)
    cmd = f"{solana_dir}solana stake-account --output json -u {url} {address}"
    res = _exec_cmd(cmd, ssh_host, ssh_key_path, timeout)
    data = {"cmd": cmd, "stdout": res.stdout, "stderr": res.stderr}
    try:
        json_res = json.loads(res.stdout)
        return Ok(StakeAccount(**json_res), data=data)
    except Exception as e:
        return Err(e, data=data)


def transfer_with_private_key_file(
    *,
    recipient: str,
    amount: Decimal,
    private_key_path: str,
    solana_dir: str = "",
    url: str = "localhost",
    ssh_host: str | None = None,
    ssh_key_path: str | None = None,
    allow_unfunded_recipient: bool = True,
    timeout: int = 60,
) -> Result[str]:
    solana_dir = _solana_dir(solana_dir)
    cmd = f"{solana_dir}solana transfer {recipient} {amount} --from {private_key_path} --fee-payer {private_key_path}"
    if allow_unfunded_recipient:
        cmd += " --allow-unfunded-recipient"
    cmd += f" -u {url} --output json"
    res = _exec_cmd(cmd, ssh_host, ssh_key_path, timeout)
    data = {"cmd": cmd, "stdout": res.stdout, "stderr": res.stderr}
    try:
        json_res = json.loads(res.stdout)
        return Ok(json_res["signature"], data=data)
    except Exception as e:
        return Err(e, data=data)


def transfer_with_private_key_str(
    *,
    recipient: str,
    amount: Decimal,
    private_key: str,
    tmp_dir_path: str,
    solana_dir: str = "",
    url: str = "localhost",
    ssh_host: str | None = None,
    ssh_key_path: str | None = None,
    timeout: int = 60,
) -> Result[str]:
    # make private_key file
    private_key_path = f"{tmp_dir_path}/solana__{random.randint(1, 10_000_000_000)}.json"
    with open(private_key_path, "w") as f:
        f.write(private_key)

    try:
        return transfer_with_private_key_file(
            recipient=recipient,
            amount=amount,
            private_key_path=private_key_path,
            solana_dir=solana_dir,
            url=url,
            ssh_host=ssh_host,
            ssh_key_path=ssh_key_path,
            timeout=timeout,
        )
    finally:
        os.remove(private_key_path)


def withdraw_from_vote_account(
    *,
    recipient: str,
    amount: Decimal | Literal["ALL"],
    vote_key_path: str,
    fee_payer_key_path: str,
    solana_dir: str = "",
    url: str = "localhost",
    ssh_host: str | None = None,
    ssh_key_path: str | None = None,
    timeout: int = 60,
) -> Result[str]:
    solana_dir = _solana_dir(solana_dir)
    cmd = f"{solana_dir}solana withdraw-from-vote-account --keypair {fee_payer_key_path} -u {url} --output json {vote_key_path} {recipient} {amount}"  # pylint: disable=line-too-long # noqa
    res = _exec_cmd(cmd, ssh_host, ssh_key_path, timeout)
    data = {"cmd": cmd, "stdout": res.stdout, "stderr": res.stderr}
    try:
        json_res = json.loads(res.stdout)
        return Ok(json_res["signature"], data=data)
    except Exception as e:
        return Err(e, data=data)


def get_validators_info(
    *,
    solana_dir: str = "",
    url: str = "localhost",
    ssh_host: str | None = None,
    ssh_key_path: str | None = None,
    timeout: int = 60,
) -> Result[list[ValidatorInfo]]:
    solana_dir = _solana_dir(solana_dir)
    cmd = f"{solana_dir}solana validator-info get --output json -u {url}"
    res = _exec_cmd(cmd, ssh_host, ssh_key_path, timeout)
    data = {"cmd": cmd, "stdout": res.stdout, "stderr": res.stderr}
    try:
        validators = []
        for v in json.loads(res.stdout):
            validators.append(
                ValidatorInfo(
                    info_address=v["infoPubkey"],
                    identity_address=v["identityPubkey"],
                    name=pydash.get(v, "info.name"),
                    keybase=pydash.get(v, "info.keybaseUsername"),
                    details=pydash.get(v, "info.details"),
                    website=pydash.get(v, "info.website"),
                ),
            )
        return Ok(validators, data=data)
    except Exception as e:
        return Err(e, data=data)


def get_vote_account_rewards(
    *,
    address: str,
    solana_dir: str = "",
    url: str = "localhost",
    ssh_host: str | None = None,
    ssh_key_path: str | None = None,
    num_rewards_epochs: int = 10,
    timeout: int = 60,
) -> Result[dict[int, float]]:
    solana_dir = _solana_dir(solana_dir)
    cmd = f"{solana_dir}solana vote-account {address} --with-rewards --num-rewards-epochs={num_rewards_epochs} -u {url}"
    cmd += " --output json 2>/dev/null"
    res = _exec_cmd(cmd, ssh_host, ssh_key_path, timeout)
    data = {"cmd": cmd, "stdout": res.stdout, "stderr": res.stderr}
    try:
        rewards: dict[int, float] = {}
        for r in reversed(json.loads(res.stdout)["epochRewards"]):
            rewards[r["epoch"]] = r["amount"] / 10**9
        return Ok(rewards, data=data)
    except Exception as e:
        return Err(e, data=data)


def get_stakes(
    *,
    vote_address: str = "",
    solana_dir: str = "",
    url: str = "localhost",
    ssh_host: str | None = None,
    ssh_key_path: str | None = None,
    timeout: int = 60,
) -> Result[list[Stake]]:
    solana_dir = _solana_dir(solana_dir)
    cmd = f"{solana_dir}solana stakes --output json -u {url} {vote_address}"
    res = _exec_cmd(cmd, ssh_host, ssh_key_path, timeout)
    data = {"stdout": res.stdout, "stderr": res.stderr}
    try:
        return Ok([Stake(**x) for x in json.loads(res.stdout)], data=data)
    except Exception as e:
        return Err(e, data=data)


def _exec_cmd(cmd: str, ssh_host: str | None, ssh_key_path: str | None, timeout: int) -> CommandResult:
    if ssh_host:
        return run_ssh_command(ssh_host, cmd, ssh_key_path, timeout=timeout)
    return run_command(cmd, timeout=timeout)


def _solana_dir(solana_dir: str) -> str:
    if solana_dir and not solana_dir.endswith("/"):
        solana_dir += "/"
    return solana_dir



from mm_std import print_console

from mm_solana.account import generate_account, get_private_key_arr_str


def run(limit: int, array: bool) -> None:
    result = {}
    for _ in range(limit):
        acc = generate_account()
        private_key = acc.private_key_base58
        if array:
            private_key = get_private_key_arr_str(acc.private_key_base58)
        result[acc.public_key] = private_key
    print_console(result, print_json=True)
